import sqlite3
import dataclasses
from datetime import datetime
import decimal

import typing
from typing import Set, Type, Any, TypeVar, List, Union, Dict, NoReturn, Optional, TYPE_CHECKING, Protocol, ClassVar
from types import TracebackType

from hoernchen.language import Column, Table, Condition, AndCondition, OrCondition, ICondition, Join, ConditionedTable, OrderedTable, JoinHow
from hoernchen.typing import Q, JoinHow, Operator, Order, StoreID, V, T
from hoernchen.column_types import COLUMN_TYPES, SQLITE_TYPES, DefaultType


def assert_never(value: NoReturn) -> NoReturn:
    assert False, f"Unhandled value: {value} ({type(value).__name__})"


def _sqlite_to_decimal(s: bytes) -> decimal.Decimal:
    return decimal.Decimal(s.decode("utf8"))


def _decimal_to_sqlite(d: decimal.Decimal) -> bytes:
    return str(d).encode("utf8")


class Store:

    def __init__(self, path: str):
        self._path = path
        self._tables: Set[str] = set()
        self._types: Set[type] = set()

        self._initialized = False
        self._in_transaction = 0

    def __enter__(self) -> "Store":
        self._begin()
        return self

    def __exit__(self, exc_type: Optional[type[BaseException]], exc_value: Optional[BaseException], exc_traceback: Optional[TracebackType]) -> None:
        if self._in_transaction:
            if exc_type is None:
                self._commit()
            else:
                self._rollback()

    def _register_dataclass(self, dc_type: Type[T]) -> None:
        self._types.add(dc_type)

    def _get_primary(self, dc_type: Type[T]) -> str:
        for field in dataclasses.fields(dc_type):
            if field.type == StoreID:
                return field.name
            if field.metadata.get("store.type", None) == "primary":
                return field.name
        return "rowid"

    def _create_table_queries(self, dc_type: Type[T]) -> list[str]:
        queries = Table.from_dataclass_type(dc_type)._to_sql_queries()
        return queries

    def _begin(self) -> None:
        if not self._initialized:
            self.create_db()
        if self._in_transaction == 0:
            self._connection.execute("BEGIN TRANSACTION;")
        self._in_transaction += 1

    def _commit(self) -> None:
        if not self._initialized:
            raise Exception("store._commit() called with uninitialized store")
        self._in_transaction -= 1
        if self._in_transaction == 0:
            self._connection.execute("COMMIT;")

    def _rollback(self) -> None:
        if not self._initialized:
            raise Exception("store.__rollback() called with uninitialized store")
        self._connection.execute("ROLLBACK;")
        self._in_transaction = 0

    def create_db(self) -> None:
        for column_type in COLUMN_TYPES.values():
            if isinstance(column_type, DefaultType):
                continue
            sqlite3.register_converter(column_type.sqlite_type, column_type.to_python)
            sqlite3.register_adapter(column_type.handles, column_type.to_sqlite)
        self._connection = sqlite3.connect(self._path, detect_types=sqlite3.PARSE_DECLTYPES, isolation_level=None)
        cur = self._connection.cursor()
        cur.execute("pragma journal_mode = WAL;")
        cur.execute("pragma synchronous = normal;")
        cur.execute("pragma temp_store = memory;")
        cur.execute("pragma mmap_size = 30000000000;")
        # cur.execute("pragma page_size = 32768;")
        self._initialized = True

        with self:
            for dc_type in self._types:
                for q in self._create_table_queries(dc_type):
                    cur.execute(q)
            for dc_type in self._types:
                self._tables.add(dc_type.__name__)

    def create_table(self, dc_type: Type[T]) -> None:
        qs = self._create_table_queries(dc_type)
        cur = self._connection.cursor()
        with self:
            for q in qs:
                cur.execute(q)
        self._tables.add(dc_type.__name__)

    def insert(self, dc: T | List[T], update: bool = False) -> None:
        if not isinstance(dc, list):
            dc = [dc]
        dc0 = dc[0]
        dc_type = type(dc0)
        tablename = dc_type.__name__

        if not self._initialized:
            self.create_db()

        if tablename not in self._tables:
            self.create_table(dc_type)

        aliases = [field.name for field in dataclasses.fields(dc_type)]
        placeholders = ["?" for field in dataclasses.fields(dc_type)]

        q = f"INSERT INTO {tablename} ({', '.join(aliases)}) VALUES ({', '.join(placeholders)})"

        if update:
            num_fillers = 1
            upsert_clause_l_fields = []
            for field in dataclasses.fields(dc_type):
                upsert_clause_l_fields.append(f"{field.name}=?")
            set_data_clause = ", ".join(upsert_clause_l_fields)

            primary = self._get_primary(dc_type)
            if primary != "rowid":
                q += f" ON CONFLICT({primary}) DO UPDATE SET {set_data_clause} "
                num_fillers += 1

            for constraint in Table.from_dataclass_type(dc_type)._constraints:
                q += f" ON CONFLICT({', '.join(constraint.columns)}) DO UPDATE SET {set_data_clause} "
                num_fillers += 1

            fillers = []
            for instance in dc:
                tup = dataclasses.astuple(instance)
                fillers.append(tuple([value for _ in range(num_fillers) for value in tup]))
        else:
            fillers = [dataclasses.astuple(instance) for instance in dc]

        q += ";"

        cur = self._connection.cursor()
        with self:
            cur.executemany(q, fillers)

    def get_single(self, dc_type: Type[T], **kwargs: Any) -> Optional[T]:
        ts = self.get(dc_type, **kwargs)
        if len(ts) == 1:
            return ts[0]
        elif len(ts) == 0:
            return None
        else:
            raise KeyError("Multiple values for query")

    def get_where(self, dc_type: Type[T], kwargs: Any = {}, order_by: Optional[str] = None, reverse: bool = False, limit: Optional[int] = None) -> List[T]:
        fields = [field.name for field in dataclasses.fields(dc_type)]
        if (order_by is not None) and (order_by not in fields):
            raise ValueError(f'OrderBy parameter must be an existing column "{fields}" but is "{order_by}".')

        tablename = dc_type.__name__

        if not self._initialized:
            self.create_db()

        if tablename not in self._tables:
            self.create_table(dc_type)

        where_clause = " and ".join([f"{key} = ?" for key in kwargs])
        if where_clause:
            where_clause = "WHERE " + where_clause
        else:
            where_clause = ""
        select_clause = f"SELECT {', '.join(fields)} from {tablename} "

        order_clause = ""
        if order_by:
            order = "DESC" if reverse else "ASC"
            order_clause = f"ORDER BY {order_by} {order}"

        limit_clause = ""
        if limit:
            assert isinstance(limit, int)
            limit_clause = f"LIMIT {limit}"

        q = " ".join([select_clause, where_clause, order_clause, limit_clause, ";"])

        cur = self._connection.cursor()
        parameters = [value for value in kwargs.values()]
        cur.execute(q, tuple(parameters))
        rows = cur.fetchall()

        result = [dc_type(*args) for args in rows]
        return result

    def get(self, dc_type: Type[T], **kwargs: Any) -> List[T]:
        return self.get_where(dc_type, kwargs)

    def _get_table_postfix(self, table_counter: Dict[str, int], name: str, increment: bool = True) -> str:
        table_count = table_counter.get(name, 0)
        if increment:
            table_counter[name] = table_count + 1
        if table_count == 0:
            table_postfix = ""
        else:
            table_postfix = str(table_count)
        return table_postfix

    def _get_value_as_str(self, tableid: str, value: Any, param_container: List[Any]) -> str:
        if isinstance(value, Column):
            return f"{tableid}.{value.name}"
        else:
            param_container.append(value)
            return "?"

    def _get_cond_str(self, tableid: str, cond: Condition, param_container: List[Union[str, int]]) -> str:
        if isinstance(cond, OrCondition):
            return "(" + " or ".join([self._get_cond_str(tableid, c, param_container) for c in cond.cs]) + ")"
        elif isinstance(cond, AndCondition):
            return "(" + " and ".join([self._get_cond_str(tableid, c, param_container) for c in cond.cs]) + ")"
        elif isinstance(cond, ICondition):
            return "NOT " + self._get_cond_str(tableid, cond.c, param_container)
        elif isinstance(cond, Condition):
            left = self._get_value_as_str(tableid, cond.c1, param_container)
            if cond.c2 is None:
                return f"{left} IS NULL"
            right = self._get_value_as_str(tableid, cond.c2, param_container)
            return f"{left} {cond.operator} {right}"
        else:
            assert_never(cond)

    def select(self, query: Union[Table, ConditionedTable, OrderedTable, Join], limit: Optional[int] = None) -> List[Q]:
        table_clauses: List[str] = []
        where_clauses: List[str] = []
        order_by = ""

        param_container: List[Union[str, int]] = []
        table_counter: Dict[str, int] = {}

        table: Optional[Table] = query

        while table is not None:
            if isinstance(table, OrderedTable):
                assert order_by == ""

                table_name = table.name
                table_postfix = self._get_table_postfix(table_counter, table_name, increment=False)
                table_id = table_name + table_postfix

                column = table.column.name if isinstance(table.column, Column) else table.column
                order_by = f"ORDER BY {table_id}.{column}" + (" ASC" if table.order == "asc" else " DESC")
                table = table.table
            elif isinstance(table, ConditionedTable):
                condition = table.condition
                table = table.table

                table_name = table.name
                table_postfix = self._get_table_postfix(table_counter, table_name, increment=False)
                table_id = table_name + table_postfix

                where_clauses.append(self._get_cond_str(table_id, condition, param_container))

            elif isinstance(table, Join):
                table_name = table.t2.name
                table_postfix = self._get_table_postfix(table_counter, table.t2.name)
                table_id = table_name + table_postfix

                inner_table_name = table.inner_name
                inner_table_postfix = self._get_table_postfix(table_counter, inner_table_name, increment=False)
                inner_table_id = inner_table_name + inner_table_postfix

                how: JoinHow = table.how
                if how == "inner":
                    join_op = "INNER JOIN"
                elif how == "left":
                    join_op = "LEFT JOIN"
                else:
                    assert_never(table.how)

                table_clauses.insert(0, f"{join_op} {table_name} {table_id} ON {inner_table_id}.{table.on1} = {table_id}.{table.on2}")

                table = table.t1
            elif isinstance(table, Table):
                dc_type = table.type

                tablename = dc_type.__name__

                if tablename not in self._tables:
                    self.create_table(dc_type)

                table_postfix = self._get_table_postfix(table_counter, tablename)
                table_id = tablename + table_postfix

                fields = [f"{table_id}.{field.name}" for field in dataclasses.fields(dc_type)]

                table_clauses.insert(0, f"SELECT {', '.join(fields)} from {tablename}{table_postfix}")
                table = None
            else:
                assert_never(table)

        select_clause = " ".join(table_clauses)
        where_clause = ("WHERE " + " and ".join(where_clauses)) if where_clauses else ""

        limit_clause = f"LIMIT {limit}" if limit else ""

        clauses = [select_clause, where_clause, order_by, limit_clause]

        q = " ".join([clause for clause in clauses if clause])

        cur = self._connection.cursor()
        cur.execute(q, param_container)
        rows = cur.fetchall()

        result = [dc_type(*args) for args in rows]
        return result

    def delete_where(self, dc_type: Type[T], kwargs: Any = {}) -> None:
        tablename = dc_type.__name__

        if not self._initialized:
            self.create_db()

        if tablename not in self._tables:
            self.create_table(dc_type)

        where_clause = " and ".join([f"{key} = ?" for key in kwargs])
        if where_clause:
            where_clause = "WHERE " + where_clause
        else:
            where_clause = ""
        delete_clause = f"DELETE FROM {tablename} "

        q = " ".join([delete_clause, where_clause, ";"])

        cur = self._connection.cursor()
        parameters = [value for value in kwargs.values()]
        cur.execute(q, tuple(parameters))
        return None



import decimal
from typing import TypeVar, Callable, Generic, Type, NewType, runtime_checkable, Protocol
import abc
import warnings
import datetime
import json


JSON = str | int | list["JSON"] | dict[str, "JSON"]

Q = TypeVar("Q")

COLUMN_TYPES = {}
SQLITE_TYPES = {}


class ColumnType(abc.ABC, Generic[Q]):
    sqlite_type: str
    handles: Type[Q]

    @abc.abstractmethod
    def to_sqlite(self, value: Q) -> bytes:
        pass

    @abc.abstractmethod
    def to_python(self, value: bytes) -> Q:
        pass

    @staticmethod
    def from_callables(python_type: Type[Q], to_sqlite_converter: Callable[[Q], bytes], sqlite_type: str, to_python_converter: Callable[[bytes], Q]) -> "ColumnType":
        return FunctionColumnType(python_type, to_sqlite_converter, sqlite_type, to_python_converter)


class FunctionColumnType(ColumnType[Q], Generic[Q]):

    def __init__(self, python_type: Type[Q], to_sqlite_converter: Callable[[Q], bytes], sqlite_type: str, to_python_converter: Callable[[bytes], Q]):
        self.handles = python_type
        self.to_sqlite_converter = to_sqlite_converter
        self.sqlite_type = sqlite_type
        self.to_python_converter = to_python_converter

    def to_sqlite(self, value: Q) -> bytes:
        return self.to_sqlite_converter(value)

    def to_python(self, value: bytes) -> Q:
        return self.to_python_converter(value)


class DefaultType(ColumnType):

    def __init__(self, python_type, sqlite_type):
        self.handles = python_type
        self.sqlite_type = sqlite_type

    def to_sqlite(self, value):
        return value

    def to_python(self, value):
        return value


def register_column_type(type: ColumnType) -> None:
    if type.handles in COLUMN_TYPES:
        warnings.warn(f"Overriding handler for python type {type.handles}")
    COLUMN_TYPES[type.handles] = type

    if type.sqlite_type in SQLITE_TYPES:
        warnings.warn(f"Overriding handler for sqlite type {type.sqlite_type}")
    SQLITE_TYPES[type.sqlite_type] = type


register_column_type(DefaultType(str, "TEXT"))
register_column_type(DefaultType(bytes, "BLOB"))
register_column_type(DefaultType(int, "INTEGER"))
register_column_type(DefaultType(float, "REAL"))


def _decimal_to_sqlite(d: decimal.Decimal) -> bytes:
    return str(d).encode("utf8")


def _sqlite_to_decimal(s: bytes) -> decimal.Decimal:
    return decimal.Decimal(s.decode("utf8"))


register_column_type(ColumnType.from_callables(decimal.Decimal, _decimal_to_sqlite, "DECIMAL", _sqlite_to_decimal))


def _bool_to_sqlite(b: bool) -> bytes:
    return b"1" if b else b"0"

def _sqlite_to_bool(s: bytes) -> bool:
    return s == b"1"


register_column_type(ColumnType.from_callables(bool, _bool_to_sqlite, "BOOL", _sqlite_to_bool))


def _date_to_sqlite(val: datetime.date) -> bytes:
    return val.isoformat().encode("utf8")


def _sqlite_to_date(val: bytes) -> datetime.date:
    return datetime.date.fromisoformat(val.decode())


register_column_type(ColumnType.from_callables(datetime.date, _date_to_sqlite, "DATE", _sqlite_to_date))


def _datetime_to_sqlite(val: datetime.datetime) -> bytes:
    return val.isoformat().encode("utf8")


def _sqlite_to_datetime(val: bytes) -> datetime.datetime:
    return datetime.datetime.fromisoformat(val.decode())


register_column_type(ColumnType.from_callables(datetime.datetime, _datetime_to_sqlite, "DATETIME", _sqlite_to_datetime))

def _json_to_sqlite(val: JSON) -> bytes:
    return json.dumps(val).encode("utf8")

def _sqlite_to_json(val: bytes) -> JSON:
    return json.loads(val.decode("utf8"))

JSONType = ColumnType.from_callables(JSON, _json_to_sqlite, "JSON", _sqlite_to_json)
JSONDictType = ColumnType.from_callables(dict, _json_to_sqlite, "JSON", _sqlite_to_json)
JSONListType = ColumnType.from_callables(list, _json_to_sqlite, "JSON", _sqlite_to_json)

COLUMN_TYPES[JSON] = JSONType
COLUMN_TYPES[dict] = JSONDictType
COLUMN_TYPES[list] = JSONListType
SQLITE_TYPES["JSON"] = JSONType



from typing import TypeVar, NewType, Literal, Union, Protocol, ClassVar, Any, Type
import dataclasses
from datetime import datetime
from decimal import Decimal

class Dataclass(Protocol):
    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]

T = TypeVar("T", bound=Dataclass)

Q = TypeVar("Q")
V = TypeVar("V")
StoreID = NewType("StoreID", int)

JoinHow = Literal["inner", "left"]
Operator = Literal["<", "<=", "=", "!=", ">=", ">", "and", "or"]
Order = Literal["asc", "desc"]



from hoernchen import col, table, Store, JSON
from dataclasses import dataclass, field
import sqlite3


@dataclass
class A:
    a: str
    b: int


@dataclass
class B:
    a: str
    b: int


def test_insert_and_get():
    s = Store(":memory:")
    a1 = A("a", 3)

    s.insert(a1)

    a2 = s.get(A, b=3)[0]

    assert a1 == a2


def test_select1():
    s = Store(":memory:")

    for i in range(10):
        a = A("a", i)
        s.insert(a)

    rs = s.select(table(A).where(col("b") > 7))
    assert len(rs) == 2


def test_join1():
    s = Store(":memory:")

    for i in range(10):
        a = A("a", i)
        s.insert(a)

    for i in range(3):
        a = A("b", i)
        s.insert(a)
    s.insert(B("a", 1))

    rs = s.select(table(A).join(table(B), on="a"))
    assert len(rs) == 10


def test_join2():
    s = Store(":memory:")

    for i in range(10):
        a = A("a", i)
        s.insert(a)

    for i in range(3):
        a = A("b", i)
        s.insert(a)
    s.insert(B("a", 1))

    rs = s.select(table(A).join(table(B), on="a", how="left").where(col("a") == None)) # noqa: E711
    assert len(rs) == 3


def test_join3():
    s = Store(":memory:")

    for i in range(10):
        a = A("a", i)
        s.insert(a)

    for i in range(3):
        a = A("b", i)
        s.insert(a)
    s.insert(B("a", 1))

    rs = s.select(table(A).order_by(col("b")).join(table(B), on="a", how="left").where(col("a") == None)) # noqa: E711

    assert rs[0] == A("b", 0)
    assert rs[1] == A("b", 1)
    assert rs[2] == A("b", 2)

    rs = s.select(table(A).order_by(col("b"), "desc").join(table(B), on="a", how="left").where(col("a") == None)) # noqa: E711

    assert rs[0] == A("b", 2)
    assert rs[1] == A("b", 1)
    assert rs[2] == A("b", 0)


def test_limit1():
    s = Store(":memory:")

    for i in range(10):
        a = A("a", i)
        s.insert(a)

    for i in range(3):
        a = A("b", i)
        s.insert(a)
    s.insert(B("a", 1))

    rs = s.select(table(A).order_by(col("b")).join(table(B), on="a", how="left").where(col("a") == None), limit=1) # noqa: E711

    assert len(rs) == 1
    assert rs[0] == A("b", 0)


@dataclass
class CBoolean:
    a: str
    b: int
    c: bool


def test_boolean():
    s = Store(":memory:")
    o1 = CBoolean("hello", 3, True)
    o2 = CBoolean("World", 4, False)
    s.insert([o1, o2])

    s1 = s.get(CBoolean, b=3)[0]
    s2 = s.get(CBoolean, b=4)[0]

    assert o1 == s1
    assert o2 == s2

    assert isinstance(s1.c, bool)


def test_update():
    s = Store(":memory:")
    o1 = CBoolean("hello", 3, True)
    o2 = CBoolean("World", 4, False)
    s.insert(o1)
    s.insert(o2, update=True)

    s1 = s.get(CBoolean, b=3)[0]
    s2 = s.get(CBoolean, b=4)[0]

    assert o1 == s1
    assert o2 == s2

    assert isinstance(s1.c, bool)


@dataclass
class D:
    a: int
    b: int


def test_get_order_by():
    s = Store(":memory:")
    l = []
    for i in range(10):
        d = D(i, 10 - i)
        l.append(d)
        s.insert(d)

    assert l == s.get_where(D, order_by="a")
    assert l[::-1] == s.get_where(D, order_by="b")
    assert l[::-1] == s.get_where(D, order_by="a", reverse=True)
    assert l == s.get_where(D, order_by="b", reverse=True)


def test_get_order_by_limit():
    s = Store(":memory:")
    l = []
    for i in range(10):
        d = D(i, 10 - i)
        l.append(d)
        s.insert(d)

    assert l[:3] == s.get_where(D, order_by="a", limit=3)
    assert l[::-1][:3] == s.get_where(D, order_by="b", limit=3)
    assert l[::-1][:3] == s.get_where(D, order_by="a", reverse=True, limit=3)
    assert l[:3] == s.get_where(D, order_by="b", reverse=True, limit=3)


def test_begin_commit_without_interaction():
    s = Store(":memory:")
    with s:
        pass


@dataclass
class E:
    a: str = field(metadata={"store.type": "unique"})
    b: str = field(metadata={"store.type": "unique"})


def test_unique():
    s = Store(":memory:")
    s.insert(E("a", "b"))
    s.insert(E("a", "c"))
    s.insert(E("a", "d"))

    try:
        s.insert(E("a", "b"))
    except sqlite3.IntegrityError:
        pass
    else:
        assert False


def test_unique_update():
    s = Store(":memory:")
    s.insert(E("a", "b"))
    s.insert(E("a", "c"))
    s.insert(E("a", "d"))

    s.insert(E("a", "b"), update=True)


@dataclass
class F:
    a: JSON
    b: str
    c: int
    d: JSON

def test_json():
    s = Store(":memory:")
    s.insert(F({"a": "asdf", "b": [1, 2, 3]}, "b_value", 3, [4, 5, 6]))

    f = s.get_single(F)

    assert f.a == {"a": "asdf", "b": [1, 2, 3]}
    assert f.b == "b_value"
    assert f.c == 3
    assert f.d == [4, 5, 6]



"""
Objects used to specify tables and subset of tables. The main class is the `Table`, representing a table of a given dataclass.
This table can be narrowed by using `.where(condition)`. A condition can be created from `Column(X)` using the operators `>, >=, <, <=, ==, !=`.
For example, the following specifies all rows of a Table X where the value y is in the range (0, 10) inclusive:

    Table(X).where((Column("y") >= 0) and (Column(y) <= 10))
"""

import typing
from typing import Union, Any, List, Literal, TypeVar, Generic, Optional, Type
import dataclasses

from hoernchen.column_types import COLUMN_TYPES, ColumnType
from hoernchen.typing import Q, JoinHow, Operator, Order, V, StoreID, T

from datetime import datetime
import decimal


def is_optional(field: dataclasses.Field) -> bool:
    for type in typing.get_args(field.type):
        if type is None:
            return True
    return False


def is_unique(field: dataclasses.Field) -> bool:
    return field.metadata.get("store.type", None) == "unique"


def get_type_from_optional(field: dataclasses.Field) -> Any:
    types = typing.get_args(field.type)
    if len(types) > 2:
        raise ValueError("Union fields are not supported (besides Optional)")
    t1, t2 = types
    if t1 is None:
        return t2
    return t1


class UniqueConstraint:

    def __init__(self, columns):
        self.columns = columns


class Table(Generic[T]):
    """
    Represents a single table in SQLite
    """

    type: Type[T]
    _name: str
    columns: List["Column"]
    _constraints: List[UniqueConstraint]

    def __init__(self, dc_type: Type[T], columns=None, constraints=None):
        self.type = dc_type
        self._name = dc_type.__name__
        self.columns = columns if columns is not None else []
        self._constraints = constraints if constraints is not None else []

    @property
    def name(self):
        return self._name

    @staticmethod
    def from_dataclass_type(dc_type: Type[T]) -> "Table":
        tablename = dc_type.__name__
        columns = []
        constraints = []

        fields_for_unique = []

        for field in dataclasses.fields(dc_type):
            optional = is_optional(field)

            if optional:
                field_type = get_type_from_optional(field)
            else:
                field_type = field.type

            if is_unique(field):
                fields_for_unique.append(field.name)

            column_type = COLUMN_TYPES.get(field_type)

            if column_type is None:
                raise KeyError(f"Type {field_type} is not supported")

            columns.append(Column(field.name, column_type, optional))
        if len(fields_for_unique) > 0:
            constraints.append(UniqueConstraint(fields_for_unique))
        return Table(dc_type, columns, constraints)

    def _to_sql_queries(self):
        queries = []
        create = "CREATE TABLE IF NOT EXISTS"
        tablename = self._name

        columns = ", ".join([column._as_sql() for column in self.columns])

        queries.append(f"{create} {tablename}({columns});")

        for constraint in self._constraints:
            column_names = constraint.columns
            index_name = f"{tablename}_unique_{'_'.join(constraint.columns)}"
            queries.append(f"CREATE UNIQUE INDEX IF NOT EXISTS {index_name} ON {tablename}({', '.join(constraint.columns)});")
        return queries

    def join(self, other: "Table", on: Union["Column", str], how: JoinHow = "inner") -> "Join":
        """Join table with another table

        Args:
            other: table to join.
            on: column to join on.
            how: How the join is evaluated. Only a subset of join methods is supported.

        Returns:
            Join object representing the joined tables
        """
        return Join(self, other, on, on, how)

    def where(self, condition: "Condition") -> "ConditionedTable":
        """Specify a subset of a table

        By default a table specifies every row in that table. With `where`, a condition can be applied, to
        only specify a subset of the table, where the condition evaluates to true.

        A `Condition` is generally created from a `Column`.

        Args:
            condition: Condition that specifies the subset of the table.

        Returns:
            Subset of a table with the condition
        """
        return ConditionedTable(self, condition)

    def order_by(self, col: "Column", order: Order = "asc") -> "OrderedTable":
        """Order table by a column in ascending or descending order

        Args:
            col: Column to order by.
            order: whether in ascending or descending order.

        Returns:
            table with order specified.
        """
        return OrderedTable(self, col, order)


class Column(Generic[V]):
    """Represents a single column of a table. Generally used to specify conditions"""

    name: str
    type: Optional[ColumnType[V]]
    optional: Optional[bool]

    def __init__(self, name: str, type: Optional[ColumnType[V]]=None, optional: Optional[bool]= None):
        self.name = name
        self.type = type
        self.optional = optional

    def __lt__(self, other: V | "Column[V]") -> "Condition":
        return Condition(self, other, "<")

    def __le__(self, other: V | "Column[V]") -> "Condition":
        return Condition(self, other, "<=")

    def __eq__(self, other: V | "Column[V]") -> "Condition":  # type: ignore
        return Condition(self, other, "=")

    def __ne__(self, other: V | "Column[V]") -> "Condition":  # type: ignore
        return Condition(self, other, "!=")

    def __ge__(self, other: V | "Column[V]") -> "Condition":
        return Condition(self, other, ">=")

    def __gt__(self, other: V | "Column[V]") -> "Condition":
        return Condition(self, other, ">")

    def __repr__(self):
        s = ["Column(\"", self.name, "\""]
        if self._type is not None:
            s.append(", ")
            s.append(self._type.sqlite_type)
        s.append(")")
        return "".join(s)

    def _as_sql(self):
        l = [self.name, " ", self.type.sqlite_type]
        if self.optional:
            l.append(" NOT NULL")
        return "".join(l)


class Condition:
    """Represents a condition to specify a subset of a table

    Conditions can be combined by using `and`, `or` or inverted by `~`.
    """

    c1: Column
    c2: Union[Any, Column]

    def __init__(self, c1: Column, c2: Union[Any, Column], operator: Operator):
        self.c1 = c1
        self.c2 = c2
        self.operator = operator

    def __and__(self, other: "Condition") -> "AndCondition":
        return AndCondition([self, other])

    def __or__(self, other: "Condition") -> "OrCondition":
        return OrCondition([self, other])

    def __invert__(self) -> "Condition":
        return ICondition(self)

    def __str__(self) -> str:
        return f"Condition({self.c1} {self.operator} {self.c2})"


class AndCondition(Condition):
    """Represents `cond1 AND cond2 AND ... condN`"""

    cs: List[Condition]

    def __init__(self, cs: List[Condition]):
        self.cs = cs

    def __and__(self, other: Condition) -> "AndCondition":
        return AndCondition(self.cs + [other])


class OrCondition(Condition):
    """Represents `cond1 OR cond2 OR ... OR condN`"""

    cs: List[Condition]

    def __init__(self, cs: List[Condition]):
        self.cs = cs

    def __or__(self, other: Condition) -> "OrCondition":
        return OrCondition(self.cs + [other])


class ICondition(Condition):
    """Represents `NOT cond`"""

    c: Condition

    def __init__(self, c: Condition):
        self.c = c

    def __invert__(self) -> Condition:
        return self.c


class Join(Table):
    """Represents `table1 JOIN table2 ON on1 = on2`"""

    t1: Table
    t2: Table
    on1: Union[Column, str]
    on2: Union[Column, str]
    how: JoinHow

    def __init__(self, t1: Table, t2: Table, on1: Union[Column, str], on2: Union[Column, str], how: JoinHow) -> None:
        self.t1 = t1
        self.t2 = t2
        self.on1 = on1
        self.on2 = on2
        self.how = how

    @property
    def name(self) -> str:
        return self.t2.name

    @property
    def inner_name(self) -> str:
        return self.t1.name


class ConditionedTable(Table):
    """Represents `table WHERE condition`"""

    table: Table
    condition: Condition

    def __init__(self, table: Table, condition: Condition):
        self.table = table
        self.condition = condition

    @property
    def name(self) -> str:
        return self.table.name

    def where(self, condition: Condition) -> "ConditionedTable":
        return ConditionedTable(self.table, self.condition & condition)


class OrderedTable(Table):
    """Represents `table ORDERED BY column`"""

    table: Table
    column: Column
    order: Order

    def __init__(self, table: Table, column: Column, order: Order):
        self.table = table
        self.column = column
        self.order = order

    @property
    def name(self) -> str:
        return self.table.name



import hoernchen.language
import dataclasses

@dataclasses.dataclass
class A:
    a: str
    b: int


def test_1():
    table = hoernchen.language.Table.from_dataclass_type(A)
    assert table.columns[0].type.handles is str
    assert table.columns[0].type.sqlite_type == "TEXT"
    assert table.columns[1].type.handles is int
    assert table.columns[1].type.sqlite_type == "INTEGER"
